;;; bbdb-hooks.el --- various additional functionality for BBDB

;; Copyright (C) 1991, 1992, 1993 Jamie Zawinski <jwz@netscape.com>.
;; Copyright (C) 2010 Roland Winkler <roland.winkler@physik.uni-erlangen.de>

;; This file is part of the Insidious Big Brother Database (aka BBDB),

;; BBDB is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 2 of the License, or
;; (at your option) any later version.

;; BBDB is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:
;;; This file provides various additional functionality for BBDB
;;; See bbdb.texinfo for documentation.

;;; This file lets you do stuff like
;;;
;;; o  automatically add some string to the notes field(s) based on the
;;;    contents of header fields of the current message
;;; o  only automatically create entries when certain header fields
;;;    are matched
;;; o  don't automatically create entries when certain header fields
;;;    are matched
;;;
;;; Read the docstrings; read the texinfo file.

(require 'bbdb)
(require 'bbdb-com)
(require 'mail-parse)

;;; RW: bad code -- do not blindly load everything!
(eval-and-compile
  (condition-case()
      (progn
        (require 'gnus)
        (require 'bbdb-gnus))
    (error nil))
  (condition-case()
      (progn
        (require 'vm)
        (require 'bbdb-vm))
    (error nil))
  (condition-case()
      (progn
        (require 'mh-show)
        (require 'bbdb-mhe))
    (error nil))
  (condition-case()
      (require 'bbdb-rmail)
    (error nil)))

(defvar rmail-buffer)
(defvar mh-show-buffer)


;;; Determining whether to create a record based on the content of the
;;; current message.

(eval-when-compile
  (defvar vm-mail-buffer)
  (defvar vm-message-pointer)
  (autoload 'widget-group-match "wid-edit")
  (autoload 'vm-start-of "vm")
  (autoload 'bbdb/vm-pop-up-bbdb-buffer "bbdb-vm"))

;; These variables both need to be enabled for gnus mailreading to
;; work right. That's probably a bug, or something.
(defcustom bbdb/mail-auto-create nil
  "If this is t, then Gnus, MH, RMAIL, and VM will automatically
create new bbdb records for people you receive mail from. If this
is a function, then it is called with no arguments to decide
whether an entry should be automatically created.
You can use this to, for example, not create records for messages
which have reached you through a particular mailing list, or to
only create records automatically if the mail has a particular
subject."
  :group 'bbdb-noticing-records
  :type '(choice (const :tag "Automatically create" t)
                 (const :tag "Prompt before creating" prompt)
                 (const :tag "Do not automatically create" nil)
                 (function :tag "Create with function" bbdb-)))

(defcustom bbdb/news-auto-create nil
  "If this is t, then Gnus will automatically create new bbdb
records for people you receive mail from.  If this is a function name
or lambda, then it is called with no arguments to decide whether an
entry should be automatically created.  You can use this to, for
example, create or not create messages which have a particular
subject.  If you want to autocreate messages based on the current
newsgroup, it's probably a better idea to set this variable to t or
nil from your `gnus-select-group-hook' instead."
  :group 'bbdb-noticing-records
  :type '(choice (const :tag "Automatically create" t)
                 (const :tag "Prompt before creating" prompt)
                 (const :tag "Do not automatically create" nil)
                 (function :tag "Create with function" bbdb-)))

;;;###autoload
(defun bbdb-display-all-recipients ()
  "Display BBDB records for all addresses of the message in this buffer.
If the records do not exist, they are generated."
  (interactive)
  ;; Alert! The MUA-specific stuff is hidden in `bbdb-header-start'.
  (let ((marker (bbdb-header-start))
        ;; Why hard-coded fields? Use `bbdb-message-headers' instead.
        (fields '("from" "sender" "to" "cc" "bcc"
                  "resent-from" "resent-to" "resent-cc" "resent-bcc"))
        addresses records record)
    (with-current-buffer (marker-buffer marker)
      (dolist (field fields)
        (goto-char marker)
        (setq addresses (append (bbdb-split (or (bbdb-message-field field)
                                                "") ",")
                                addresses))))
    (dolist (address addresses)
      (setq record (bbdb-annotate-message-address address t))
      (if record (push record records)))
    (bbdb-display-records (sort records 'bbdb-record-lessp))))

;;;###autoload
(defun bbdb-header-start ()
  "Returns a marker at the beginning of the header block of the current
message.  This will not necessarily be in the current buffer."
  (cond ((memq major-mode
	       '(gnus-group-mode gnus-summary-mode gnus-article-mode))
	 (set-buffer (or gnus-original-article-buffer
                         gnus-article-buffer))
	 (point-min-marker))
        ((memq major-mode '(vm-presentation-mode vm-mode vm-summary-mode))
         (if vm-mail-buffer (set-buffer vm-mail-buffer))
         (vm-start-of (car vm-message-pointer)))
        ((memq major-mode '(rmail-mode rmail-summary-mode))
         (if (and (boundp 'rmail-buffer) rmail-buffer)
             (set-buffer rmail-buffer))
         (point-min-marker))
        ;; MH-E clause added by knabe.
        ((eq major-mode 'mh-folder-mode)
         (mh-show)
         (set-buffer mh-show-buffer)
         (point-min-marker))
        (t (point-min-marker))))

;;;###autoload
(defun bbdb-message-field (field-name)
  "Given the name of a field (like \"Subject\") this returns the value of
that field in the current message, or nil.  This works whether you're in
Gnus, Rmail, or VM.  This works on multi-line fields, but if more than
one field of the same name is present, only the last is returned.  It is
expected that the current buffer has a message in it, and (point) is at the
beginning of the message headers."
  ;; we can't special-case VM here to use its cache, because the cache has
  ;; divided real-names from addresses; the actual From: and Subject: fields
  ;; exist only in the message.
  (save-excursion
    (if (memq major-mode
              '(gnus-summary-mode gnus-article-mode gnus-tree-mode))
        (progn
          (set-buffer (get-buffer gnus-original-article-buffer))
          (goto-char (point-min))))
    (setq field-name (concat (regexp-quote field-name) "[ \t]*:[ \t]*"))
    (let ((case-fold-search t)
          done)
      (while (not (or done
              (looking-at "\n") ; we're at BOL
              (eobp)))
        (if (looking-at field-name)
        (progn
          (goto-char (match-end 0))
          (setq done (buffer-substring (point)
                       (progn (end-of-line) (point))))
          (while (looking-at "\n[ \t]")
            (setq done (concat done " "
                 (buffer-substring (match-end 0)
                   (progn (end-of-line 2) (point))))))))
        (forward-line 1))
      (and done
	   (mail-decode-encoded-word-string done)))))

(defcustom bbdb-accept-messages-alist '()
  "An alist describing which messages to automatically create BBDB
records for.  This only works if `bbdb/news-auto-create' or
`bbdb/mail-auto-create' (or both) is `bbdb-ignore-most-messages'.
The format of this alist is
   (( HEADER-NAME . REGEXP ) ... )
for example,
   ((\"From\" . \"@.*\\.maximegalon\\.edu\")
    (\"Subject\" . \"time travel\"))
will cause BBDB entries to be made only for messages sent by people at
Maximegalon U., or (that's *or*) people posting about time travel.

See also `bbdb-ignore-messages-alist', which has the opposite effect."
  :group 'bbdb-noticing-records
  :type '(repeat (cons
          (string :tag "Header name")
          (regexp :tag "Regex to match on header value"))))


(defcustom bbdb-ignore-messages-alist '()
  "An alist describing which messages *not* to automatically create
BBDB records for.  This only works if `bbdb/news-auto-create' or
`bbdb/mail-auto-create' (or both) is `bbdb-ignore-some-messages'.
The format of this alist is
   (( HEADER-NAME . REGEXP ) ... )
for example,
   ((\"From\" . \"mailer-daemon\")
    (\"To\" . \"mailing-list-1\\\\|mailing-list-2\")
    (\"CC\" . \"mailing-list-1\\\\|mailing-list-2\"))
will cause BBDB entries to not be made for messages from any mailer daemon,
or messages sent to or CCed to either of two mailing lists.

See also `bbdb-accept-messages-alist', which has the opposite effect."
  :group 'bbdb-noticing-records
  :type '(repeat (cons
          (string :tag "Header name")
          (regexp :tag "Regex to match on header value"))))


;;;###autoload
(defun bbdb-ignore-most-messages (&optional invert-sense)
  "For use as the value of `bbdb/news-auto-create' or `bbdb/mail-auto-create'.
This will automatically create BBDB entries for messages which match
the `bbdb-accept-messages-alist' (which see) and *no* others."
  ;; don't need to optimize this to check the cache, because if
  ;; bbdb/*-update-record uses the cache, this won't be called.
  (let ((rest (if invert-sense
          bbdb-ignore-messages-alist
          bbdb-accept-messages-alist))
        (case-fold-search t)
        (buffer (current-buffer))
        (marker (bbdb-header-start))
        field regexp fieldval done)
    (set-buffer (marker-buffer marker))
    (save-restriction
      (widen)
      (while (and rest (not done))
        (goto-char marker)
        (setq field (car (car rest))
              regexp (cdr (car rest))
              fieldval (bbdb-message-field field))
        (if (and fieldval (string-match regexp fieldval))
            (setq done t))
        (setq rest (cdr rest))))
    (set-buffer buffer)
    (if invert-sense
        (not done)
      done)))

;;; Provided by Bill Carpenter.
(defvar bbdb-ignore-selected-messages-confirmation nil
  "If bbdb-ignore-selected-messages-hook is used as an auto-create-hook, this
variable governs whether you are prompted for creation of BBDB entries.")

;;;###autoload
(defun bbdb-ignore-selected-messages-hook ()
  "For use as a `bbdb/news-auto-create' or `bbdb/mail-auto-create'.
This will automatically create BBDB entries for messages based on a
combination of `bbdb-ignore-messages-alist' and
`bbdb-accept-messages-alist'.  It first looks at the SOME list.  If
that doesn't disqualify a message, then it looks at the MOST list.  If
that qualifies the message, the record is auto-created, but a
confirmation is conditionally sought, based on the value of
`bbdb-ignore-selected-messages-confirmation'."
  (if (bbdb-ignore-some-messages)
      ;; wasn't ruled out
      (if (bbdb-ignore-most-messages)
          ;; was ruled in
          (if bbdb-ignore-selected-messages-confirmation
              (let ((case-fold-search t)
                    (marker (bbdb-header-start))
                    record-exists from)
                (with-current-buffer (marker-buffer marker)
                  (save-restriction
                    (widen)
                    (goto-char marker)
                    (setq from (bbdb-message-field "FROM"))))
                (setq record-exists (bbdb-annotate-message-address from))
                (or record-exists
                    (y-or-n-p (concat "Create BBDB record from " from "? "))))
            ;; no confirmation desired so let it be
            t))))

;;;###autoload
(defun bbdb-ignore-some-messages ()
  "For use as a `bbdb/news-auto-create' or `bbdb/mail-auto-create'.
This will automatically create BBDB entries for messages which do *not*
match the `bbdb-ignore-messages-alist' (which see)."
  (bbdb-ignore-most-messages t))


;;; Automatically add to the notes field based on the current message.

(define-widget 'bbdb-alist-with-header 'group
  "My group"
  :match 'bbdb-alist-with-header-match
  :value-to-internal (lambda (widget value)
                       (if value (list (car value) (cdr value))))
  :value-to-external (lambda (widget value)
                       (if value (append (list (car value)) (cadr value)))))

(defun bbdb-alist-with-header-match (widget value)
  (widget-group-match widget
                      (widget-apply widget :value-to-internal value)))

(defcustom bbdb-auto-notes-alist nil
  "An alist which lets you have certain pieces of text automatically added
to the BBDB record representing the sender of the current message based on
the subject or other header fields.  This only works if `bbdb-notice-hooks'
contains `bbdb-auto-notes-hook'.  The format of this alist is

   ((HEADER-NAME [HEADER-CLASS-LIST]
       (REGEXP . STRING) ... )
      ... )
for example,
   ((\"To\" (\"-vm@\" . \"VM mailing list\"))
    (\"Subject\" (\"sprocket\" . \"mail about sprockets\")
               (\"you bonehead\" . \"called me a bonehead\")))

will cause the text \"VM mailing list\" to be added to the notes field of
the record corresponding to anyone you get mail from via one of the VM
mailing lists.  If, that is, `bbdb/mail-auto-create' returns such that the
record would have been created, or the record already existed.

A HEADER-CLASS-LIST is optional and by default actions will be performed only
for records of senders of a message.  However, by giving an list of classes
specified in `bbdb-message-headers'.  Actions will then only be
performed if the currently processed mail is of a class listed in
HEADER-CLASS-LIST.  HEADER-CLASS-LIST might also be an alist with elements
of the form (CLASS . HEADER) which allows actions only when the current
address matches one of the elemets.

The format of elements of this list may also be
       (REGEXP FIELD-NAME STRING)
or
       (REGEXP FIELD-NAME STRING REPLACE-P)
instead of
       (REGEXP . STRING)

meaning add the given string to the named field.  The field-name may not
be name, address, phone, or mail (builtin fields) but must be either ``notes,''
``organization,'' or the name of a user-defined note-field.
       (\"pattern\" . \"string to add\")
is equivalent to
       (\"pattern\" notes \"string to add\")

STRING can contain \\& or \\N escapes like in function
`replace-match'.  For example, to automatically add the contents of the
\"organization\" field of a message to the \"organization\" field of a BBDB
record, you can use this:

        (\"Organization\" (\".*\" organization \"\\\\&\"))

\(Note you need two \\ to get a single \\ into a lisp string literal.\)

If STRING is an integer N, the N'th matching subexpression is used, so
the above example could be written more efficiently as

        (\"Organization\" (\".*\" organization 0))

If STRING is neither a string or an integer, it should be a function, which
will be called with the contents of the field.  The result of that function
call is used as the field value (the returned value must be a string.)

If REPLACE-P is t, the string replaces the old contents instead of
being appended to it.

If multiple clauses match the message, all of the corresponding strings
will be added.

This works for news as well.  You might want to arrange for this to have
a different value when in mail as when in news.

See also variables `bbdb-auto-notes-ignore' and `bbdb-auto-notes-ignore-all'."
  :group 'bbdb-noticing-records
  :type '(repeat
          (bbdb-alist-with-header
           (string :tag "Header name")
           (repeat (choice
                    (cons :tag "Address Class"
                          (repeat (choice
                                   (const sender)
                                   (const recipients))))
                    (cons :tag "Value Pair"
                          (regexp :tag "Regexp to match on header value")
                          (string :tag "String for notes if regexp matches"))
                    (list :tag "Replacement list"
                          (regexp :tag "Regexp to match on header value")
                          (choice :tag "Record field"
                                  (const notes :tag "Notes")
                                  (const organization :tag "Organization")
                                  (symbol :tag "Other"))
                          (choice :tag "Regexp match"
                                  (string :tag "Replacement string")
                                  (integer :tag "Subexpression match")
                                  (function :tag "Callback Function"))
                          (choice :tag "Replace previous contents"
                                  (const :tag "No" nil)
                                  (const :tag "Yes" t))))))))

(defcustom bbdb-auto-notes-ignore nil
  "Alist of headers and regexps to ignore in `bbdb-auto-notes-hook'.
Each element looks like

    (HEADER . REGEXP)

For example,

    (\"Organization\" . \"^Gatewayed from\\\\\|^Source only\")

would exclude the phony `Organization:' headers in GNU mailing-lists
gatewayed to gnu.* newsgroups.  Note that this exclusion applies only
to a single field, not to the entire message.  For that, use the variable
`bbdb-auto-notes-ignore-all'."
  :group 'bbdb-noticing-records
  :type '(repeat (cons
          (string :tag "Header name")
          (regexp :tag "Regexp to match on header value"))))

(defcustom bbdb-auto-notes-ignore-all nil
  "Alist of headers and regexps which cause the entire message to be ignored
in `bbdb-auto-notes-hook'.  Each element looks like

    (HEADER . REGEXP)

For example,

    (\"From\" . \"BLAT\\\\.COM\")

would exclude any notes recording for message coming from BLAT.COM.
Note that this is different from `bbdb-auto-notes-ignore', which applies
only to a particular header field, rather than the entire message."
  :group 'bbdb-noticing-records
  :type '(repeat (cons
          (string :tag "Header name")
          (regexp :tag "Regexp to match on header value"))))


;;;###autoload
(defun bbdb-auto-notes-hook (record)
  "For use as a `bbdb-notice-hooks'.  This might automatically add some text
to the notes field of the BBDB record corresponding to the current record
based on the header of the current message.  See the documentation for
the variables `bbdb-auto-notes-alist' and `bbdb-auto-notes-ignore'."
  ;; This could stand to be faster...
  ;; could optimize this to check the cache, and noop if this record is
  ;; cached for any other message, but that's probably not the right thing.
  (unless bbdb-read-only
   (let ((rest bbdb-auto-notes-alist)
         (ignore-all bbdb-auto-notes-ignore-all)
         (case-fold-search t)
         (buffer (current-buffer))
         (marker (bbdb-header-start))
         ignore
         field pairs fieldval  ; do all bindings here for speed
         regexp string notes-field-name notes
         replace-p)
    (set-buffer (marker-buffer marker))
    (save-restriction
      (widen)
      (goto-char marker)
      (if (and (setq fieldval (bbdb-message-field "From"))
               (string-match bbdb-user-mail-names fieldval))
          ;; Don't do anything if this message is from us.  Note that we have
          ;; to look at the message instead of the record, because the record
          ;; will be of the recipient of the message if it is from us.
          nil
        ;; check the ignore-all pattern
        (while (and ignore-all (not ignore))
          (goto-char marker)
          (setq field (car (car ignore-all))
                regexp (cdr (car ignore-all))
                fieldval (bbdb-message-field field))
          (if (and fieldval
                   (string-match regexp fieldval))
              (setq ignore t)
            (setq ignore-all (cdr ignore-all))))

        (unless ignore          ; ignore-all matched
         (while rest ; while there are still clauses in the auto-notes alist
          (goto-char marker)
          (setq field (car (car rest))  ; name of header, e.g., "Subject"
                pairs (cdr (car rest))  ; (REGEXP . STRING) or
                                        ; (REGEXP FIELD-NAME STRING) or
                                        ; (REGEXP FIELD-NAME STRING REPLACE-P)
                fieldval (bbdb-message-field field)) ; e.g., Subject line
          (when fieldval
            ;; we perform the auto notes stuff only for sender of a message
            ;; or if explicitly requested
            (if (or (symbolp (caar pairs)) (listp (caar pairs)))
                (if (or (memq (nth 3 bbdb-update-records-address) (car pairs))
                        (and (assoc (nth 3 bbdb-update-records-address) (car pairs))
                             (string= (nth 2 bbdb-update-records-address)
                                      (cdr (assoc (nth 3 bbdb-update-records-address)
                                                  (car pairs))))))
                    (setq pairs (cdr pairs))
                  (setq pairs nil))
              (if (not (and (eq 'sender (nth 3 bbdb-update-records-address))
                            (string-match "From" (nth 2 bbdb-update-records-address))))
                  (setq pairs nil)))

            ;; now handle the remaining pairs
            (while pairs
              (setq regexp (car (car pairs))
                    string (cdr (car pairs)))
              (if (consp string) ; not just the (REGEXP . STRING) format
                  (setq notes-field-name (car string)
                        replace-p (nth 2 string) ; perhaps nil
                        string (nth 1 string))
                  ;; else it's simple (REGEXP . STRING)
                  (setq notes-field-name 'notes
                        replace-p nil))
              (setq notes (bbdb-record-note record notes-field-name))
              (let ((did-match
                     (and (string-match regexp fieldval)
                          ;; make sure it is not to be ignored
                          (let ((re (cdr (assoc field
                                                bbdb-auto-notes-ignore))))
                            (if re
                                (not (string-match re fieldval))
                                t)))))
                ;; An integer as STRING is an index into match-data:
                ;; A function as STRING calls the function on fieldval:
                (if did-match
                    (setq string
                          (cond ((integerp string) ; backward compat
                                 (substring fieldval
                                            (match-beginning string)
                                            (match-end string)))
                                ((stringp string)
                                 (bbdb-auto-expand-newtext fieldval string))
                                (t
                                 (goto-char marker)
                                 (let ((s (funcall string fieldval)))
                                   (or (stringp s)
                                       (null s)
                                       (error "%s returned %s: not a string"
                                              string s))
                                   s)))))
                ;; need expanded version of STRING here:
                (if (and did-match
                         string ; A function as STRING may return nil
                         (not (and notes
                                   ;; check that STRING is not already
                                   ;; present in the NOTES field
                                   (string-match
                                    (regexp-quote string)
                                    notes))))
                    (if replace-p
                        ;; replace old contents of field with STRING
                        (progn
                          (unless bbdb-silent
                            (if (eq notes-field-name 'notes)
                                (message "Replacing with note \"%s\"" string)
                              (message "Replacing field \"%s\" with \"%s\""
                                       notes-field-name string)))
                          (bbdb-record-set-note record notes-field-name string)
                          (bbdb-maybe-update-display record))
                        ;; add STRING to old contents, don't replace
                      (unless bbdb-silent
                        (if (eq notes-field-name 'notes)
                            (message "Adding note \"%s\"" string)
                          (message "Adding \"%s\" to field \"%s\""
                                   string notes-field-name)))
                      (bbdb-annotate-notes record string notes-field-name))))
              (setq pairs (cdr pairs))))
          (setq rest (cdr rest))))))
    (set-buffer buffer))))

(defun bbdb-auto-expand-newtext (string newtext)
  ;; Expand \& and \1..\9 (referring to STRING) in NEWTEXT.
  ;; Note that in Emacs 18 match data are clipped to current buffer
  ;; size...so the buffer had better not be smaller than STRING (arrrrggggh!!)
  (let ((pos 0)
    (len (length newtext))
    (expanded-newtext ""))
    (while (< pos len)
      (setq expanded-newtext
        (concat expanded-newtext
            (let ((c (aref newtext pos)))
              (if (= ?\\ c)
              (cond ((= ?\& (setq c (aref newtext
                              (setq pos (1+ pos)))))
                 (substring string
                        (match-beginning 0)
                        (match-end 0)))
                ((and (>= c ?1)
                      (<= c ?9))
                 ;; return empty string if N'th
                 ;; sub-regexp did not match:
                 (let ((n (- c ?0)))
                   (if (match-beginning n)
                       (substring string
                          (match-beginning n)
                          (match-end n))
                     "")))
                (t (char-to-string c)))
            (char-to-string c)))))
      (setq pos (1+ pos)))
    expanded-newtext))


;;; I use this as the value of `bbdb-canonicalize-mail-function'.
;;; It is provided as an example for you to customize.

(defcustom bbdb-canonical-hosts
  (mapconcat 'regexp-quote
         '("cs.cmu.edu" "ri.cmu.edu" "edrc.cmu.edu" "andrew.cmu.edu"
           "mcom.com" "netscape.com" "cenatls.cena.dgac.fr"
           "cenaath.cena.dgac.fr" "irit.fr" "enseeiht.fr" "inria.fr"
           "cs.uiuc.edu" "xemacs.org")
         "\\|")
  "Certain sites have a single mail-host; for example, all mail originating
at hosts whose names end in \".cs.cmu.edu\" can (and probably should) be
addressed to \"user@cs.cmu.edu\" instead.  This variable lists other hosts
which behave the same way."
  :group 'bbdb
  :type '(regexp :tag "Regexp matching sites"))

;;;###autoload
(defun bbdb-sample-canonicalize-mail-function (address)
  (cond
   ;;
   ;; rewrite mail-drop hosts.
   ;;
   ((string-match
     (concat "\\`\\([^@%!]+@\\).*\\.\\(" bbdb-canonical-hosts "\\)\\'")
     address)
    (concat (match-string 1 address) (match-string 2 address)))
   ;;
   ;; Here at Lucid, our workstation names sometimes get into our mail
   ;; addresses in the form "jwz%thalidomide@lucid.com" (instead of simply
   ;; "jwz@lucid.com").  This removes the workstation name.
   ;;
   ((string-match "\\`\\([^@%!]+\\)%[^@%!.]+@\\(lucid\\.com\\)\\'" address)
    (concat (match-string 1 address) "@" (match-string 2 address)))
   ;;
   ;; Another way that our local mailer is misconfigured: sometimes addresses
   ;; which should look like "user@some.outside.host" end up looking like
   ;; "user%some.outside.host" or even "user%some.outside.host@lucid.com"
   ;; instead.  This rule rewrites it into the original form.
   ;;
   ((string-match "\\`\\([^@%]+\\)%\\([^@%!]+\\)\\(@lucid\\.com\\)?\\'" address)
    (concat (match-string 1 address) "@" (match-string 2 address)))
   ;;
   ;; Sometimes I see addresses like "foobar.com!user@foobar.com".
   ;; That's totally redundant, so this rewrites it as "user@foobar.com".
   ;;
   ((string-match "\\`\\([^@%!]+\\)!\\([^@%!]+[@%]\\1\\)\\'" address)
    (match-string 2 address))
   ;;
   ;; Sometimes I see addresses like "foobar.com!user".  Turn it around.
   ;;
   ((string-match "\\`\\([^@%!.]+\\.[^@%!]+\\)!\\([^@%]+\\)\\'" address)
    (concat (match-string 2 address) "@" (match-string 1 address)))
   ;;
   ;; The mailer at hplb.hpl.hp.com tends to puke all over addresses which
   ;; pass through mailing lists which are maintained there: it turns normal
   ;; addresses like "user@foo.com" into "user%foo.com@hplb.hpl.hp.com".
   ;; This reverses it.  (I actually could have combined this rule with
   ;; the similar lucid.com rule above, but then the regexp would have been
   ;; more than 80 characters long...)
   ;;
   ((string-match "\\`\\([^@!]+\\)%\\([^@%!]+\\)@hplb\\.hpl\\.hp\\.com\\'"
          address)
    (concat (match-string 1 address) "@" (match-string 2 address)))
   ;;
   ;; Another local mail-configuration botch: sometimes mail shows up
   ;; with addresses like "user@workstation", where "workstation" is a
   ;; local machine name.  That should really be "user" or "user@netscape.com".
   ;; (I'm told this one is due to a bug in SunOS 4.1.1 sendmail.)
   ;;
   ((string-match "\\`\\([^@%!]+\\)[@%][^@%!.]+\\'" address)
    (match-string 1 address))
   ;;
   ;; Sometimes I see addresses like "foo%somewhere%uunet.uu.net@somewhere.else".
   ;; This is silly, because I know that I can send mail to uunet directly.
   ;;
   ((string-match ".%uunet\\.uu\\.net@[^@%!]+\\'" address)
    (concat (substring address 0 (+ (match-beginning 0) 1)) "@UUNET.UU.NET"))
   ;;
   ;; Otherwise, leave it as it is.  Returning a string EQ to the one passed
   ;; in tells BBDB that we're done.
   ;;
   (t address)))


;;; Here's another approach; sometimes one gets mail from foo@bar.baz.com,
;;; and then later gets mail from foo@baz.com.  At this point, one would
;;; like to delete the bar.baz.com address, since the baz.com address is
;;; obviously superior.  See also var `bbdb-canonicalize-redundant-mails'.
;;;
;;; Turn this on with
;;;   (add-hook 'bbdb-change-hooks 'bbdb-delete-redundant-mails)

(defun bbdb-delete-redundant-mails (record)
  "Deletes redundant mail addresses.
For use as a value of `bbdb-change-hooks'.  See `bbdb-mail-redundant-p'."
  (let* ((mails (bbdb-record-mail record))
     (rest mails)
     mail new redundant)
    (while rest
      (setq mail (car rest))
      (if (bbdb-mail-redundant-p mail mails)
      (setq redundant (cons mail redundant))
    (setq new (cons mail new)))
      (setq rest (cdr rest)))
    (cond (redundant
       (message "Deleting redundant mails %s..."
                (bbdb-join ", " (nreverse redundant)))
       (setq new (nreverse new))
       (bbdb-record-set-mail record new)
       t))))


;;;###autoload
(defun bbdb-force-record-create ()
  "Force automatic creation of a BBDB records for the current message.
You might add this to the reply hook of your MUA in order to automatically
get records added for those people you reply to."
  (interactive)
  (let ((bbdb/mail-auto-create t)
        (bbdb/news-auto-create t)
        bbdb-message-caching-enabled
        (bbdb/gnus-update-records-p 'annotate)
        (bbdb/rmail-update-records-p 'annotate)
        (bbdb/mhe-update-records-p 'annotate)
        (bbdb/vm-update-records-p 'annotate))
    (save-excursion
      (cond ((member major-mode '(vm-mode vm-virtual-mode vm-summary-mode
                                          vm-presentation-mode))
             (bbdb/vm-pop-up-bbdb-buffer))
            ((member major-mode '(gnus-summary-mode gnus-article-mode
                                                    gnus-tree-mode))
             (bbdb/gnus-pop-up-bbdb-buffer))
            ((member major-mode '(rmail-mode rmail-summary-mode))
             (bbdb/rmail-pop-up-bbdb-buffer))
            ((member major-mode '(mhe-mode mhe-summary-mode mh-folder-mode))
             (bbdb/mh-pop-up-bbdb-buffer))))))

(provide 'bbdb-hooks)
