;;; bbdb-rmail.el --- BBDB interface to Rmail

;; Copyright (C) 1991, 1992 Jamie Zawinski <jwz@netscape.com>.
;; Copyright (C) 2010 Roland Winkler <roland.winkler@physik.uni-erlangen.de>

;; This file is part of the Insidious Big Brother Database (aka BBDB),

;; BBDB is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 2 of the License, or
;; (at your option) any later version.

;; BBDB is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:
;;; This file contains the BBDB interface to Rmail.
;;; See bbdb.texinfo for documentation.

(require 'bbdb)
(require 'bbdb-com)
(require 'rmail)
(require 'rmailsum)
(require 'mailheader)

(defcustom bbdb/rmail-update-records-p
  (lambda () (if (bbdb/rmail-new-flag rmail-buffer) 'annotate-q 'search))
  "Controls how `bbdb/rmail-update-records' processes mail addresses.
Set this to an expression which evaluates to 'search, 'annotate. or nil.
When set to 'annotate mail addresses will be fed to
`bbdb-annotate-message-address' in order to update existing records or create
new ones.  A value of 'search will search just for existing records having
the right mail.  A value of nil will not do anything.

The default is to annotate only new messages."
  :group 'bbdb-mua-rmail
  :type '(choice (const :tag "annotate all messages"
                        annotate)
                 (const :tag "annotate no messages"
                        search)
                 (const :tag "annotate only new messages"
                        (if (lambda () (bbdb/rmail-new-flag rmail-buffer) 'annotate-q 'search)))
                 (sexp  :tag "user defined")))

(defun bbdb/rmail-new-flag (buf)
  "Returns t if the current message in buffer BUF is new."
  (rmail-message-labels-p rmail-current-message ", ?\\(unseen\\),"))

(defun bbdb/rmail-header-content (header-field buf)
  "Pull HEADER-FIELD out of BUF's mail header.
BUF is actually the rmail buffer from which the current message should
be extracted."
  (with-current-buffer buf
    (if (fboundp 'rmail-get-header)    ; Emacs 23
        (rmail-get-header header-field)
      (save-restriction
        (with-no-warnings (rmail-narrow-to-non-pruned-header))
        (let ((headers (mail-header-extract))
              (header (intern-soft (downcase header-field))))
          (mail-header header headers))))))

;;;###autoload
(defun bbdb/rmail-update-records (&optional create-p)
  "Return the records corresponding to the current RMAIL message,
creating or modifying them as necessary.  A record will be created if
CREATE-P is non-nil."
  (if (and (boundp 'rmail-buffer) rmail-buffer)
      (set-buffer rmail-buffer)
    (error "Not in an rmail buffer"))
  (when rmail-current-message
    (let (records)
      (unless create-p
        (setq create-p
              (if (functionp bbdb/rmail-update-records-p)
                  (funcall bbdb/rmail-update-records-p)
                bbdb/rmail-update-records-p)))
      ;; ignore cache if we may be creating a record, since the cache
      ;; may otherwise tell us that the user didn't want a record for
      ;; this person.
      (unless (member create-p '(annotate annotate-q))
        (setq records (bbdb-message-get-cache rmail-current-message)))
      (unless records
        (setq records (bbdb-update-records
                       (bbdb-get-address-components
                        user-mail-address
                        'bbdb/rmail-header-content '(rmail-buffer))
                       create-p))
        (bbdb-message-set-cache rmail-current-message records))
      (if bbdb-message-all-addresses
          records
        (if records (list (car records)))))))

;;;###autoload
(defun bbdb/rmail-annotate-sender (string &optional replace)
  "Add a line to the end of the Notes field of the BBDB record
corresponding to the sender of this message.  If REPLACE is non-nil,
replace the existing notes entry (if any)."
  (interactive (progn (bbdb-editable) (list (read-string "Comments: "))))
  (if (and (boundp 'rmail-buffer) rmail-buffer)
      (set-buffer rmail-buffer))
  (dolist (record (bbdb/rmail-update-records))
    (bbdb-annotate-notes record string 'notes replace)))

(defun bbdb/rmail-edit-notes (&optional field)
  "Edit the notes FIELD of the BBDB record corresponding to the sender
of this message.
If called interactively, FIELD defaults to 'notes. With a prefix arg,
ask interactively for FIELD."
  (interactive (list (unless current-prefix-arg 'notes)))
  (let ((records (bbdb/rmail-update-records)))
    (bbdb-display-records records)
    (dolist (record records)
      (bbdb-record-edit-notes record field t))))

;; ;;;###autoload
;; (defun bbdb/rmail-show-sender ()
;;   "Display the contents of the BBDB for the sender of this message."
;;   (interactive)
;;   (if (and (boundp 'rmail-buffer) rmail-buffer)
;;       (set-buffer rmail-buffer))
;;   (let ((record (bbdb/rmail-update-record)))
;;     (if record
;;         (bbdb-display-records (list record))
;;         (error "unperson"))))

(defun bbdb/rmail-pop-up-bbdb-buffer (&optional create-p)
  "Make the *BBDB* buffer be displayed along with the RMAIL window(s).
Displays the records corresponding to the sender respectively
recipients of the current message.
See `bbdb-message-headers' and `bbdb-message-all-addresses'
for configuration of what is being displayed."
  (if bbdb-message-pop-up
      (let ((bbdb-silent-internal t)
            (records (bbdb/rmail-update-records create-p)))
        (if records
            (bbdb-display-records-internal
             records nil nil nil
             (lambda (window)
               (let ((buffer (current-buffer)))
                 (set-buffer (window-buffer window))
                 (prog1 (eq major-mode 'rmail-mode)
                   (set-buffer buffer)))))
          ;; If there are no records, empty the BBDB window.
          (bbdb-undisplay-records)))))

;;;###autoload
(defun bbdb-insinuate-rmail ()
  "Call this function to hook BBDB into RMAIL."
  (define-key rmail-mode-map ":" 'bbdb/rmail-show-sender)
  (define-key rmail-mode-map ";" 'bbdb/rmail-edit-notes)
  (define-key rmail-summary-mode-map ":" 'bbdb/rmail-show-sender)
  (define-key rmail-summary-mode-map ";" 'bbdb/rmail-edit-notes)

  (add-hook 'rmail-show-message-hook 'bbdb/rmail-pop-up-bbdb-buffer)

  ;; We must patch into rmail-only-expunge to clear the cache, since
  ;; expunging a message invalidates the cache (which is based on
  ;; message numbers).
  (defadvice rmail-only-expunge (before bbdb/rmail-only-expunge)
    "Invalidate BBDB cache before expunging."
    (setq bbdb-message-cache nil))

  ;; Same for undigestifying.
  (or (fboundp 'undigestify-rmail-message)
      (autoload 'undigestify-rmail-message "undigest" nil t))
  (if (eq (car-safe (symbol-function 'undigestify-rmail-message)) 'autoload)
      (load (nth 1 (symbol-function 'undigestify-rmail-message))))
  (defadvice undigestify-rmail-message (before bbdb/undigestify-rmail-message)
    "Invalidate BBDB cache before undigestifying."
    (setq bbdb-message-cache nil)))

(provide 'bbdb-rmail)
